
import std/strutils
import std/hashes
import std/tables
import std/locks
import std/times
import std/json
import std/uri
import std/os

import arr_ref
import loadable

export arr_ref

import sugar

type CacheResult* = object
    ok*: bool
    err*: string
    orig_data*, cached_data*: SliceMem[byte]

type CacheCallback* = proc(res: CacheResult)

type CacheEntry = object
    cached: string
    modified: Time
    size: int
    settings: string
    # TODO: flag to exclude from cache?

var cache_lock: Lock
var cached_files {.guard: cache_lock.}: Table[string, CacheEntry]
var cache_dir {.guard: cache_lock.}: string
template CACHE_DB_PATH: string = cache_dir & "cache.db"

initLock(cache_lock)

proc uri_to_cache_file(uri: string): string =
    var path = uri
    try:
        let parsed = parseUri(uri)
        if parsed.scheme.len > 0:
            path = parsed.path & parsed.query & parsed.anchor
    except: discard
    
    for c in path:
        if c.isAlphaNumeric:
            result.add c
        else:
            result.add '_'

    result = result.strip(chars = {'_'}) & "_" & $uri.hash

var last_dir {.threadVar.}: string

proc setCacheDir*(dir: string) =
    assert dir != "", "A directory must be supplied."
    if last_dir == dir:
        return
    last_dir = dir
    # TODO: make dir?
    assert dir.dirExists
    withLock cache_lock:
        cache_dir = dir.normalizePathEnd(trailingSep = true) # add /
    # load database
    try:
        withLock cache_lock:
            cached_files = readFile(CACHE_DB_PATH).parseJson.to(typeof(cached_files))
    except:
        echo "Could not load cache database"

proc storeCachedURI*(cache_key: string, cached_data: openArray[byte], dir = "") {.gcsafe.} =
    assert cache_key != "", "cache_key can't be empty"
    var dir = dir.normalizePathEnd(trailingSep = true)
    withLock cache_lock:
        {.cast(gcsafe).}:
            assert cache_key in cached_files, "You need to call getCachedURI/Memory first."
            assert cache_dir != "", "Cache directory is not configured. Call setCacheDir()."
            if dir == "":
                dir = cache_dir
    assert cached_data.len != 0, "Can't use an empty file as cache"
    # if it exists, it's overwritten so we don't need to check if there's an
    # existing entry
    let file_name = uri_to_cache_file(cache_key)
    var f = open(dir & file_name, fmWrite)
    defer: f.close()
    if f.writeBuffer(cached_data[0].addr, cached_data.len) != cached_data.len:
        echo "Error: Could not write cache file"
        return
    withLock cache_lock:
        {.cast(gcsafe).}:
            cached_files[cache_key].cached = file_name
            # save database
            writeFile CACHE_DB_PATH, $(%cached_files)

proc getCachedURI*[T](uri: string, cache_key: string, settings: T, callback: CacheCallback, use_cache = true) =
    if not use_cache:
        loadUri uri, proc(res: FetchResult) =
            callback CacheResult(ok: res.ok, err: res.err, orig_data: res.data)
        return

    var dir: string
    var entry: CacheEntry
    withLock cache_lock:
        assert cache_dir != "", "Cache directory is not configured. Call setCacheDir()."
        dir = cache_dir
        entry = cached_files.getOrDefault(cache_key)
    let settings = when compiles($settings):
        $settings
    elif compiles(%settings):
        $(%settings)
    else:
        {.error: "Can't convert settings to string or JSON".}
    
    # TODO: do we need a way to ensure we're not requesting the file multiple timmes?

    loadUri uri, headers_only = true, onload_func = proc(res: FetchResult) =
        template fallback() =
            # Make cache entry without the actual file, to be added later
            withLock cache_lock:
                cached_files[cache_key] = CacheEntry(
                    modified: res.modified,
                    size: res.size,
                    settings: $settings,
                )
            loadUri uri, proc(res: FetchResult) =
                callback CacheResult(ok: res.ok, err: res.err, orig_data: res.data)
        
        if not res.ok:
            when not defined(myouLoadCacheOnFailure):
                # pass through error
                callback CacheResult(ok: false, err: res.err)
            else:
                # give cached file instead
                loadUri dir & entry.cached, proc(res: FetchResult) =
                    callback CacheResult(ok: res.ok, err: res.err, cached_data: res.data)
        # NOTE: time 0 means it may not be invalidated on change
        if entry.cached == "" or
            entry.modified < res.modified or
                entry.size != res.size or
                entry.settings != $settings:
            fallback()
            return
        
        loadUri dir & entry.cached, proc(res: FetchResult) =
            if res.ok:
                callback CacheResult(ok: true, err: res.err, cached_data: res.data)
            else:
                echo "Not found: ", dir & entry.cached
                echo "Missing cache file, using original file: ", uri
                fallback()

proc getCachedFromMemory*[T](data: SliceMem[byte], cache_key: string, settings: T, callback: CacheCallback, use_cache = true) =
    if not (use_cache and cache_key != ""):
        callback CacheResult(ok: true, orig_data: data)
        return

    var dir: string
    var entry: CacheEntry
    withLock cache_lock:
        assert cache_dir != "", "Cache directory is not configured. Call setCacheDir()."
        dir = cache_dir
        entry = cached_files.getOrDefault(cache_key)
    let settings = when compiles($settings):
        $settings
    elif compiles(%settings):
        $(%settings)
    else:
        {.error: "Can't convert settings to string or JSON".}
    
    template fallback() =
        echo "folbac"
        # Make cache entry without the actual file, to be added later
        withLock cache_lock:
            cached_files[cache_key] = CacheEntry(
                size: data.len,
                settings: $settings,
                # TODO: hash
            )
        callback CacheResult(ok: true, orig_data: data)
    
    if entry.cached == "" or
            entry.size != data.len or
            entry.settings != $settings:
                # TODO: hash
        dump entry.cached
        dump (entry.size, data.len)
        dump (entry.settings, $settings)
        fallback()
        return
    
    loadUri dir & entry.cached, proc(res: FetchResult) =
        if res.ok:
            callback CacheResult(ok: true, err: res.err, cached_data: res.data)
        else:
            echo "Not found: ", dir & entry.cached
            echo "Missing cache file, using original file: ", cache_key
            fallback()


# TODO: solution for partial files (mipmaps)?
