# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import ../types
import ../graphics/ubo
import ../util
import std/strformat
import std/strutils
import std/tables

proc newShapeKeyModifier*(engine: MyouEngine, count: int): VertexModifier =

    let float_count = count.align4

    result.get_code = proc(v: seq[Varying]): VertexModifierCodeLines =
        result.uniform_lines = @[
            "layout(std140) uniform ShapeKeys {",
            &"    vec4 shape_values[{float_count div 4}];",
            "};",
            # 256/255 = 1.0039216 = 2.0078433 / 2.0
            # which helps us remap 0 to -1 and 255 to 1.
            """
                vec3 unpack(float f){
                    float x = f;
                    float y = fract(abs(x)*128.0);
                    float z = fract(y*256.0);
                    y = y * 2.0078433 - 1.0;
                    z = z * 2.0078433 - 1.0;
                    return vec3(x,y,z);
                }
            """.dedent,
        ]
        var body = @[
            # Equivalent to /= 127.0, and roughly to normalize byte normals
            "normal *= 0.007874;",
            "float relf = 0.0;",
            "vec3 co0 = co.xyz, orig_n = normal;",
        ]
        for i in 0 ..< count:
            let i4 = i div 4
            body &= @[
                &"co += vec4(shape{i}.xyz, 0.0) * shape_values[{i4}][{i and 3}];",
                &"float shape{i}use = min(1.0, length(shape{i}.xyz) * 1000.0);",
                &"relf += abs(shape_values[{i4}][{i and 3}]) * shape{i}use;",
            ]
        # Interpolating normals instead of re-calculating them is wrong
        # But it's fast, completely unnoticeable in most cases,
        # and better than just not changing them (as many engines do)
        # body &= "normal *= clamp(1.0 - relf, 0.0, 1.0);"
        body &= "vec3 snormal = normal * 0.001;"
        body &= "#define DEBUG1 0.001"
        body &= "#define DEBUG2 0.001"
        body &= "#define DEBUG3 0.001"
        for i in 0 ..< count:
            let i4 = i div 4
            # I substract 0.001 to avoid using noise near 0 which makes normals
            # change in places where a shape key is not affecting
            body &= &"""
                snormal += unpack(shape{i}.w) * max(0.0, shape_values[{i4}][{i and 3}] * shape{i}use)
            ;"""
        body &= "normal = mix(normal, snormal, clamp(relf, 0.0, 1.0));"

        result.body_lines = body
    
    let ubo = engine.renderer.newUBO("ShapeKeys", float32, float_count)
    result.ubo = ubo
    
    result.update = proc(m: Mesh) =
        var data = ubo.storage(float32)
        var i = 0
        for k,v in m.shape_keys:
            data[i] = v
            i.inc
        ubo.unbind()
        ubo.update()
    
